from .shapely_impl import *


__all__ = [
    'Shape',
    'EMPTY',
    'FULL',
    'Region',
    'SingleShape',
    'SimplePolygon',
    'ComplexPolygon',
    'MultiShape',
    'MultiSimplePolygon',
    'MultiComplexPolygon',
    'create_polygon',
    'create_triangle',
    'create_regular_polygon',
    'create_sector',
    'ConvertMulti2SingleException',
    'load',
    'loads',
    'loadb',
]



from abc import ABC
from typing import TypeVar, Generic, TextIO, BinaryIO
from typing import Tuple, Union, List, Any, Iterable


# 定义一个类型变量
T = TypeVar('T', bound='ShapeInterface')

# 坐标类型
Position = Union[str, complex, Tuple[float, float]]


class ShapeInterface(ABC, Generic[T]):
    """
    平面图形计算库
     -- >> 仅计算【 平面图形 】，不考虑三维图像
     -- >> 仅计算【 有面积的区域 】，凡没有面积的，诸如”点“、”线“，一律直接滤除
     -- >> 仅计算【 有限表示图形 】，诸如分形图形这种强依赖于数学构造而不可能绘制的图形，不予讨论

    # 一个推荐的 C++ 实现方案: clipper

    自定义接口, 约定以下操作:
    1. 坐标运算:
    位移运算 -> offset, 参数 (x, y) 或 x+yj, 占用 左加运算符, 左减运算符 [+, -]
    缩放运算 -> scale, 参数 float, 占用 左乘运算符, 左除运算符 [*, /]     -->> 默认情况下以原点为基准做缩放变换，可以输入关键词 'center'，以轮廓中心为基准做缩放变换
    * 旋转运算 -> rotate, 参数 float, 占用 左幂运算符 [**], 逆时针为正
    * 镜像运算 -> flip_x, flip_y, flip, 不占用运算符, 其中:
        flip_x: 入参 x0, 表示以 x=x0 为对称轴做镜像
        flip_y: 入参 y0, 表示以 y=y0 为对称轴做镜像
        flip: 入参 degree, origin, 其中 origin == (x0, y0), 表示以 tan(degree) = (y - y0) / (x - x0) 为对称轴做镜像
    2. 集合论运算:
    交集 -> inter, 占用 与运算符 [&]
    并集 -> union, 占用 或运算符 [|]
    判定相交 -> is_joint, 不占用运算符
    判定覆盖 -> if_contain, 不占用运算符
    * 异集 -> diff, 取双方互不隶属的元素组成新集合, 占用 异或运算符 [^]
    * 补集 -> comp, 交换集合内外元素, 占用 按位反运算符 [~]
    * 合集 -> merge, 在相交的情况下，令 other: T 附着至 self, 占用 左移运算符 [<<]
    * 差集 -> remove, 将 other: T 的元素从 self 中移除, 占用 右移运算符 [>>]
    3. 形态学运算:
    凸包络 -> convex, 不占用运算符
    密接矩形 -> mini_rect, 不占用运算符
    矩形域 -> region, 不占用运算符
    轮廓化简 -> simplify, 参数 int, 表示轮廓点密度, 不占用运算符
    轮廓平滑 -> smooth, 参数 int, 表示平滑程度, 不占用运算符
    4. 形态学属性
    形心 -> center, 返回 (x, y), 不占用运算符
    面积 -> area, 返回 float, 不占用运算符
    周长 -> perimeter, 返回 float, 不占用运算符
    矩形边界 -> bounds, 返回 (l, u, r, d), 不占用运算符
    5. 工程属性
    表示方式 -> reversed, 当前轮廓是正表示的还是负表示的(无穷远不属于轮廓 -> 正表示), 返回 bool, 不占用运算符, 受 补集运算支配
    正形 -> outer, 返回外轮廓(正表示, FULL 除外), 占用 自正运算符 [+]
    负形 -> inner, 返回内轮廓(正表示, FULL 除外), 占用 自反运算符 [-]
    真值 -> bool, 返回当前图形是否存在轮廓表示（应当为真，除非 Shape 是 empty、full 或等效为空的异常情况）
    6. 轮廓的分解
    内解 -> sep_in, 返回 正形和负形, 不占用运算符
    外解 -> sep_out, 多类型返回类型数组, 单类型返回自身构成的数组, 占用 迭代运算符 [iter]
    点解 -> sep_p, 返回类型结构数组, 参照具体类型说明, 不占用运算符
    7. 不支持的运算符:
    [
        加减乘除幂的右运算符 +, -, *, /, **
        取模运算符 abs, 取整运算符 round,
        比较运算符 <, <=, ==, !=, >=, >,
        赋值运算符 =, +=, -=, *=, /=, **=, //=, &=, |=, ^=, %=, <<=, >>=
        取余运算符 %, 整除运算符 //, 带余除法运算符 divmod
    ]
    8. 特殊形及操作
    空形 -> EMPTY, 同 [], 正形: EMPTY, 负形: FULL,内解: [], [FULL], 外解: [], 不可点解
    全形 -> FULL, 同 [FULL], 正形: FULL, 负形: EMPTY,内解: [FULL], [], 外解: [FULL], 不可点解
    9. 读写
    写出 -> dump/dumps, 将图像写出为文件或字符串
    加载 -> load/loads, 从文件或字符串中加载图像
    """

    __slots__ = ()

    def offset(self, vector: Position) -> T:
        # 位移变换(原地的)
        raise NotImplementedError

    def scale(self, ratio: float, origin: Position = 0j) -> T:
        # 缩放变换(原地的)
        raise NotImplementedError

    def rotate(self, degree: float, origin: Position = 0j) -> T:
        # 旋转变换(原地的)
        raise NotImplementedError

    def flip_x(self, x0: float) -> T:
        # 镜像变换, 对称轴 x=x0 (原地的)
        raise NotImplementedError

    def flip_y(self, y0: float) -> T:
        # 镜像变换, 对称轴 y=y0 (原地的)
        raise NotImplementedError

    def flip(self, degree: float, origin: Position) -> T:
        # 镜像变换, 对称轴 x/a + y/b = 1 (原地的)
        raise NotImplementedError

    def inter(self, other: T) -> T:
        # 集合论 交集运算 (创建新对象)
        raise NotImplementedError

    def union(self, other: T) -> T:
        # 集合论 并集运算 (创建新对象)
        raise NotImplementedError

    def is_joint(self, other: T) -> bool:
        # 集合论 判断相交
        raise NotImplementedError

    def if_contain(self, other: T) -> bool:
        # 集合论 判断覆盖
        raise NotImplementedError

    def diff(self, other: T) -> T:
        # 集合论 异集运算 (创建新对象)
        raise NotImplementedError

    def comp(self) -> T:
        # 集合论 补集运算 (创建新对象)
        raise NotImplementedError

    def merge(self, others: Iterable[T]) -> T:
        # 集合论 合集运算 (创建新对象)
        raise NotImplementedError

    def remove(self, other: T) -> T:
        # 集合论 差集运算 (创建新对象)
        raise NotImplementedError

    def simplify(self, tolerance: float) -> T:
        # 形态学 轮廓简化 (创建新对象)
        raise NotImplementedError

    def smooth(self, distance: float) -> T:
        # 形态学 轮廓平滑 (创建新对象)
        raise NotImplementedError

    def buffer(self, distance: float) -> T:
        # 形态学 轮廓腐蚀或膨胀 (创建新对象)
        raise NotImplementedError

    @property
    def convex(self) -> T:
        # 形态学 凸包络 (创建新对象)
        raise NotImplementedError

    @property
    def mini_rect(self) -> T:
        # 形态学 最小外接矩形 (创建新对象)
        raise NotImplementedError

    @property
    def region(self) -> T:
        # 形态学 最小坐标矩形 (创建新对象)
        raise NotImplementedError

    @property
    def center(self) -> Tuple[int, int]:
        # 形态学 形心
        raise NotImplementedError

    @property
    def area(self) -> float:
        # 形态学 面积
        raise NotImplementedError

    @property
    def perimeter(self) -> float:
        # 形态学 周长
        raise NotImplementedError

    @property
    def bounds(self) -> Tuple[int, int, int, int]:
        # 形态学 坐标边界
        raise NotImplementedError

    @property
    def reversed(self) -> bool:
        # 工程学 返回翻转表示标志的布尔值
        raise NotImplementedError

    @property
    def outer(self) -> T:
        # 工程学 由外轮廓构成的多边形
        raise NotImplementedError

    @property
    def inner(self) -> T:
        # 工程学 由内轮廓构成的多边形, 若无内轮廓, 返回 EMPTY
        raise NotImplementedError

    def sep_in(self) -> Tuple[T, T]:
        # 工程学 将多边形分解为内轮廓多边形与外轮廓多边形
        raise NotImplementedError

    def sep_out(self) -> List[T]:
        # 工程学 将多边形分解为多边形数组,其中每个多边形只含一个复连通区域
        raise NotImplementedError

    def sep_p(self) -> Any:
        # 工程学 将多边形分解为坐标点列,这个点列只使用 python 原生的 list,tuple, 便于数据的传输和迁移
        raise NotImplementedError

    def copy(self) -> T:
        # 工程学 创建新对象
        raise NotImplementedError

    def standard(self) -> T:
        # 工程学 轮廓规范化 (创建新对象)
        raise NotImplementedError

    # 下列运算符均创建新对象
    def __add__(self, pos: Union[complex, Tuple[float, float]]) -> T:
        return self.copy().offset(pos)

    def __sub__(self, pos: Union[complex, Tuple[float, float]]) -> T:
        if isinstance(pos, complex):
            return self.copy().offset(-pos)
        else:
            return self.copy().offset((-pos[0], -pos[1]))

    def __mul__(self, ratio: float) -> T:
        return self.copy().scale(ratio)

    def __truediv__(self, ratio: float) -> T:
        assert ratio != 0, 'Could not div by 0!'
        return self.copy().scale(1 / ratio)

    def __pow__(self, degree: float) -> T:
        return self.copy().rotate(degree)

    def __and__(self, other: T) -> T:
        return self.copy().inter(other)

    def __or__(self, other: T) -> T:
        return self.copy().union(other)

    def __xor__(self, other: T) -> T:
        return self.copy().diff(other)

    def __invert__(self) -> T:
        return self.copy().comp()

    def __lshift__(self, other: T) -> T:
        return self.copy().merge(other)

    def __rshift__(self, other: T) -> T:
        return self.copy().remove(other)

    def __pos__(self) -> T:
        return self.copy().outer

    def __neg__(self) -> T:
        return self.copy().inner

    def __iter__(self) -> Iterable[T]:
        return iter(self.sep_out())

    # 下列运算符均就地的修改对象
    def __iadd__(self, pos: Union[complex, Tuple[float, float]]) -> T:
        return self.offset(pos)

    def __isub__(self, pos: Union[complex, Tuple[float, float]]) -> T:
        if isinstance(pos, complex):
            return self.offset(-pos)
        else:
            return self.offset((-pos[0], -pos[1]))

    def __imul__(self, ratio: float) -> T:
        return self.scale(ratio)

    def __itruediv__(self, ratio: float) -> T:
        assert ratio != 0, 'Could not div by 0!'
        return self.scale(1 / ratio)

    def __ipow__(self, degree: float) -> T:
        return self.rotate(degree)

    def __iand__(self, other: T) -> T:
        return self.inter(other)

    def __ior__(self, other: T) -> T:
        return self.union(other)

    def __ixor__(self, other: T) -> T:
        return self.diff(other)

    def __ilshift__(self, other: T) -> T:
        return self.merge(other)

    def __irshift__(self, other: T) -> T:
        return self.remove(other)

    def dump(self, f: TextIO) -> None:
        f.write(self.dumps())

    def dumps(self) -> str:
        raise NotImplementedError

    def dumpb(self, f: BinaryIO) -> None:
        raise NotImplementedError



import importlib
my_modules = {
    'Logger': '.logger',
    'TimerManager': '.timer',
    'Timer': '.timer',
    'Queue': '.multiprocess',
    'Closed': '.multiprocess',
    'Process': '.multiprocess',
    'QueueMessageException': '.multiprocess',
    'JassorJsonEncoder': '.json_encoder',
    'Merger': '.merger',
    'random_colors': '.color',
    'random_rainbow_curves': '.color',
    'plot': '.jassor_plot_lib',
    'plots': '.jassor_plot_lib',
}


def __getattr__(name):
    if name in my_modules:
        module = importlib.import_module(my_modules[name], __package__)
        return getattr(module, name)
    else:
        raise ModuleNotFoundError('你脑瓜子被驴踢啦，我才没有这个方法呢，你给我检查清楚了再 import')


__all__ = list(my_modules)



from typing import Any
import json


class JassorJsonEncoder(json.JSONEncoder):
    # 只支持 dict、list、tuple、str、number 且 dict 的 key 必须是 str
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # print(*args)
        # print(list(kwargs.items()))
        self._indent = kwargs['indent'] or 4
        self._count = 0

    def encode(self, obj: Any) -> str:
        if isinstance(obj, (list, tuple)) and all(map(lambda x: not isinstance(x, (list, tuple, dict)), obj)) or \
                isinstance(obj, dict) and all(map(lambda v: not isinstance(v, (list, tuple, dict)), obj.values())):
            json_str = json.dumps(obj)
        elif isinstance(obj, dict):
            end = '\n' + ' ' * self._count + '}'
            self._count += self._indent
            head = '{\n' + ' ' * self._count
            transformer = lambda kv: f'"{str(kv[0])}": {self.encode(kv[1])}'
            content = f',\n{" " * self._count}'.join(map(transformer, obj.items()))
            self._count -= self._indent
            json_str = head + content + end
        elif isinstance(obj, (list, tuple)):
            end = '\n' + ' ' * self._count + ']'
            self._count += self._indent
            head = '[\n' + ' ' * self._count
            content = f',\n{" " * self._count}'.join(map(self.encode, obj))
            self._count -= self._indent
            json_str = head + content + end
        else:
            json_str = json.dumps(obj)
        return json_str

    def iterencode(self, o, _one_shot=False):
        return self.encode(o)



import json
import types
from typing import Any, List
from io import TextIOWrapper
from pathlib import Path
from threading import Condition
from typing import TextIO
import numpy as np
import multiprocessing
from typing import TypeVar

class JassorJsonEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs) -> None: ...
    def encode(self, obj: Any) -> str: ...
    def iterencode(self, o, _one_shot: bool = False): ...

class IOWrapper:
    def __init__(self, write_func: callable, flush_func: callable, close_func: callable): ...

class Logger:
    STEP: int
    DEBUG: int
    INFO: int
    WARNING: int
    def __init__(self, start: int = 0, indentation: int = 1, file: TextIO | IOWrapper | TextIOWrapper | str | Path = ..., con: Condition = None, level: int = 2) -> None: ...
    def close(self) -> None: ...
    def track(self, message: str, prefix: str = ''): ...
    def step(self, message: str): ...
    def debug(self, message: str): ...
    def info(self, message: str): ...
    def warn(self, message: str): ...
    def tab(self): ...
    def __getitem__(self, item: str): ...
    def __enter__(self): ...
    def __exit__(self, exc_type: type | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None): ...

class Merger:
    def __init__(self, temp: np.ndarray | tuple[int, ...], kernel: np.ndarray | tuple[int, ...], dtype: type = ..., steep: float = 4, eps: float = 1e-17) -> None: ...
    @staticmethod
    def get_kernel(shape: tuple[int, ...], steep: float) -> np.ndarray: ...
    def set(self, patch: np.ndarray, grid: tuple[int | None, ...]) -> None: ...
    def tail(self) -> np.ndarray: ...

V = TypeVar('V')

class Process(multiprocessing.Process):
    def __init__(self, target: callable, args=()) -> None: ...
    def start(self) -> None: ...
    def run(self) -> None: ...

class Queue:
    def __init__(self, maxsize: int = 0) -> None: ...
    def ready(self) -> bool: ...
    def is_end(self) -> bool: ...
    def count(self) -> int: ...
    def push(self, item: V) -> None: ...
    def top(self) -> V: ...
    def pop(self) -> V: ...
    def message(self) -> str: ...
    def end(self, flag: bool = False, message: str = '') -> None: ...

class Closed(BaseException):
    def __init__(self, message: str) -> None: ...

class QueueMessageException(BaseException):
    def __init__(self, message: str) -> None: ...
class TimerManager:
    register: dict[str, dict[str, float]]
    def __init__(self, container: str | dict[str, float] = None) -> None: ...
    def __getitem__(self, item: str): ...
    @property
    def costs(self) -> dict[str, float]: ...
    @staticmethod
    def stamp(t: float) -> str: ...
    @staticmethod
    def register_container(key: str, container: dict[str, float]): ...

class Timer:
    def __init__(self, key: str, container: str | dict[str, float]) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type: type | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None): ...
    def __call__(self, func: callable): ...

def random_colors(n: int, start: tuple[int, int, int] = (47, 81, 177), step: tuple[int, int, int] = (73, 119, 31), rand: tuple[int, int, int] = (10, 10, 10)) -> list[tuple[int, int, int]]: ...
def random_rainbow_curves(shape: tuple[int, int, int], s: int = 117, k: int = 7, c: int = 50): ...
def plot(item: Any, title: str = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000): ...
def plots(items: List[Any], titles: List[str] = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000): ...



from typing import Union, Tuple
import numpy as np


class Merger(object):
    def __init__(
            self,
            temp: Union[np.ndarray, Tuple[int, ...]],
            kernel: Union[np.ndarray, Tuple[int, ...]],
            dtype: type = np.float32,
            steep: float = 4,
            eps: float = 1e-17,
    ):
        """
        一个相对通用的融合器，参数可以直接提供一个 ndarray，也可以提供一个用于生成 ndarray 的形状
        :param temp: 融合模板，用来缓存融合结果
        :param kernel: 融合系数，用来实现有差融合，需要广播的维度请设为 1
        :param dtype: 以参数方式生成 ndarray 时的数据类型
        :param steep: 以参数方式生成 kernel 时，kernel 形状的描述符，kernel 越大，中心区域价值占比越高
        """
        if isinstance(temp, tuple):
            temp = np.zeros(shape=temp, dtype=dtype)
        if isinstance(kernel, tuple):
            kernel = self.get_kernel(shape=kernel, steep=steep).astype(dtype)
        assert len(temp.shape) == len(kernel.shape), 'attention : temp.shape.length eq kernel.shape.length eq patch.shape.length'
        self._temp = temp
        self._kernel = kernel
        # kernel 中维度数为 1 视为需要广播，helper 中置为 1 即可
        # 否则视为需要贴片，按照 temp 的维度数来设置
        helper_shape = tuple(1 if k == 1 else t for k, t in zip(kernel.shape, temp.shape))
        self._helper = np.zeros(shape=helper_shape, dtype=dtype) + eps

    @staticmethod
    def get_kernel(shape: Tuple[int, ...], steep: float) -> np.ndarray:
        # 第一步：获得一个对称区间，x = np.arange(w).astype(np.float32) - (w - 1)/2
        # 第二步：计算高斯函数，x = exp(-1/2 (x/sigma) ** 2)，其中 sigma = size / steep
        # 第三步：乘到 kernel 里
        kernel = np.asarray([1], dtype=np.float32)
        for i, size in enumerate(shape):
            sigma = size / steep
            x = np.arange(size).astype(np.float32) - (size - 1) / 2
            x = np.exp(-1 / 2 * (x / sigma) ** 2)
            x /= x.mean()
            kernel = kernel[..., None] @ x[(None,) * (i + 1)]
        return np.ascontiguousarray(kernel[0])

    def set(self, patch: np.ndarray, grid: Tuple[Union[int, None], ...]) -> None:
        """
        贴图方法，patch 表示待贴图块，grid 表示图块坐标
        :param patch: 需确保 dtype 与 kernel 一致，shape 与 kernel 在广播形式下一致
        :param grid: 整数坐标，无需计算（需要广播的维度，或锁定满填充的维度）的维度置为 None
        """
        # 计算图
        patch = patch * self._kernel
        # 将 grid 坐标 limit 在坐标范围之内，构成贴片
        # 针对 temp 的计算
        temp_grid = tuple(
            slice(None) if g is None else slice(max(0, g), min(s, max(0, g + k)))
            for s, k, g in zip(self._temp.shape, self._kernel.shape, grid)
        )
        # # 针对 helper 的计算
        # help_grid = tuple(
        #     (0, 1) if g is None else (max(0, g), min(s, g + k))
        #     for s, k, g in zip(self._temp.shape, patch.shape, grid)
        # )
        # 针对 patch 的计算
        patch_grid = tuple(
            slice(None) if g is None else slice(max(0, -g), min(k, max(0, s - g)))
            for s, k, g in zip(self._temp.shape, self._kernel.shape, grid)
        )
        # # 针对 kernel 的计算
        # kernel_grid = tuple(
        #     (0, 1) if g is None else (max(0, -g), min(k, s - g))
        #     for s, k, g in zip(self._temp.shape, patch.shape, grid)
        # )
        # 以上面的计算为基础执行贴图方法
        self._temp[temp_grid] += patch[patch_grid]
        self._helper[temp_grid] += self._kernel[patch_grid]

        # for (x, y), target in zip(grids, targets.cpu()):
        #     temp_left = max(0, x)
        #     temp_up = max(0, y)
        #     temp_right = min(self.w, x + self.ksize)
        #     temp_down = min(self.h, y + self.ksize)
        #     patch_left = max(0, -x)
        #     patch_up = max(0, -y)
        #     patch_right = min(self.ksize, self.w - x)
        #     patch_down = min(self.ksize, self.h - y)
        #     # print(x, y, target[patch_up: patch_down, patch_left: patch_right, :].shape, self.target[temp_up: temp_down, temp_left: temp_right, :].shape)
        #     self.target[temp_up: temp_down, temp_left: temp_right, :] += target[patch_up: patch_down, patch_left: patch_right, :]
        #     self._helper[temp_up: temp_down, temp_left: temp_right, :] += self.k[patch_up: patch_down, patch_left: patch_right, :]

    def tail(self) -> np.ndarray:
        return self._temp / self._helper



from typing import TypeVar
from queue import Empty, Full
import multiprocessing


V = TypeVar('V')


class Process(multiprocessing.Process):
    def __init__(self, target: callable, args=()):
        super(Process, self).__init__(target=target, args=args)

    def start(self) -> None:
        super(Process, self).start()

    def run(self) -> None:
        super(Process, self).run()


class Queue(object):
    """
    这是一个简易的队列封装设计，占用基础类型 None 和 True
    建议通过字典格式传输控制信息和数据
    ready: 阻塞直到队列获得新元素，队列结束时返回 False
    is_end: 非阻塞判断队列是否结束时使用
    count: 调查队列运行情况时使用，不可靠，不建议用作条件判断
    push: 向队列中压入元素的方法
    pop: 从队列中取出元素的方法
    top: 查看队列头部元素的方法（请注意，该方法在线程/进程条件下表现不同，不同线程共享同一 top，而不同进程占据不同 top）
    end: 结束队列（目前不支持从底层结构上结束队列，要求队列结束时任务也跟着结束）
    """
    def __init__(self, maxsize: int = 0):
        self._top = True
        self._con = multiprocessing.Condition()
        self._que = multiprocessing.Queue(maxsize=maxsize)
        # 状态： 0, 1, 2
        # 0: 异常中止
        # 1: 正常运行
        # 2: 正常中止
        self._flag = multiprocessing.Value('i', 1)
        self._message = multiprocessing.Array('c', b' ' * 1000)

    def ready(self) -> bool:
        # 阻塞判定方法，当且仅当队列下一个元素已就绪时返回真
        return self._flag.value and self.top() is not None

    def is_end(self) -> bool:
        # 非阻塞判定方法，当队列已声明为不可用状态，或队列元素已取完并声明结束时返回真
        return not(self._flag.value and self._top is not None)

    def count(self) -> int:
        # 非阻塞方法，返回当前队列内的元素数量（不保证准确）
        return self._flag.value and (self._que.qsize() + (self._top is not True) - (self._flag.value & 0b10))

    def push(self, item: V) -> None:
        # 阻塞方法，向队列内推送值
        while self._flag.value == 1:
            # 正常运行状态，正常使用
            try:
                self._que.put(item, timeout=0.5)
                break
            except Full:
                pass
        else:
            if self._flag.value == 2:
                # 正常结束状态
                return
            # 异常结束状态，直接报错
            raise Closed(self.message())

    def top(self) -> V:
        # 阻塞独占方法，查看队列顶端元素值
        with self._con:
            while self._flag.value:
                # 正常运行状态，正常使用
                try:
                    if self._top is True:
                        self._top = self._que.get(timeout=0.5)
                    break
                except Empty:
                    if self._flag.value == 2:
                        # 正常结束状态
                        self._top = None
                        return None
            else:
                # 异常结束状态，直接报错
                raise Closed(self.message())
        return self._top

    def pop(self) -> V:
        # 阻塞独占方法，从队列中取值
        with self._con:
            item = self.top()
            if item is not None:
                self._top = True
        return item

    def message(self) -> str:
        return self._message.value.decode()

    def end(self, flag: bool = False, message: str = '') -> None:
        """
        结束队列，含两种结束方式，正常结束或异常结束
        flag == False: 正常结束，当前队列用完后不再接受新数据
        flag == True: 异常结束，立刻丢弃队列内元素，不再接受新数据
        """
        if flag:
            # 异常结束
            self._flag.value = 0
        else:
            # 正常结束
            self._flag.value = 2
            try:
                self._que.put(None, timeout=0.1)
                self._que.put(None, timeout=0.2)
            except Full:
                pass
        self._message.value = message.encode()


class Closed(BaseException):
    def __init__(self, message: str):
        super(Closed, self).__init__(message)


class QueueMessageException(BaseException):
    def __init__(self, message: str):
        super(QueueMessageException, self).__init__(message)



import math
from typing import List, Any, Iterable, Tuple
import shapely
from PIL import Image
import matplotlib
from matplotlib.axes import Axes
matplotlib.use('TkAgg')     # 好像只有这个支持
import matplotlib.pyplot as plt
import cv2
import numpy as np
from shapely.geometry.base import BaseGeometry


def plot(item: Any, title: str = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000):
    title = title or ''
    fig = plt.figure()
    ax = fig.add_subplot(111)
    _plot_item(ax, item)
    ax.set_title(str(title))
    ax.set_aspect('equal')

    fig.canvas.manager.set_window_title(window_name)
    plt.tight_layout()
    if not save_to:
        plt.show()
    else:
        plt.savefig(save_to, dpi=dpi)
    plt.close(fig=fig)


def plots(items: List[Any], titles: List[str] = None, window_name: str = 'jassor_plot', save_to: str = None, dpi: int = 1000):
    n = len(items)
    titles = (titles or []) + [''] * n
    # 计算行列数量
    row = int(n ** 0.5)
    col = (n + row - 1) // row

    fig, axs = plt.subplots(row, col)
    axs = [axs] if row == col == 1 else axs.flatten()
    for ax in axs[n:]:
        fig.delaxes(ax)

    for ax, item, title in zip(axs, items, titles):
        _plot_item(ax, item)
        ax.set_title(str(title))
        ax.set_aspect('equal')

    fig.canvas.manager.set_window_title(window_name)
    plt.tight_layout()
    if not save_to:
        plt.show(block=True)
    else:
        plt.savefig(save_to, dpi=dpi)
    plt.close(fig=fig)


def _plot_item(ax: Axes, item: Any) -> Any:

    # 支持三大类显示方式：
    # 图像显示：np.array, torch.Tensor, PIL.Image, str（file_path）
    # 点列显示：[[(x1, y1), (x2, y2), ...]], [[x1, x2, ...], [y1, y2, ...]]
    # 轮廓显示：shapely.geometry

    if item is None:
        return ax.imshow(_draw_empty(f'item is None'))

    if isinstance(item, np.ndarray):
        # ndarray 必须是 (h, w, (1, 3, 4)) 或 (h, w)
        if len(item.shape) not in (2, 3) or min(item.shape) == 0:
            return ax.imshow(_draw_empty(f'only (h, w, c) or (h, w) are supported, got {item.shape}'))
        if len(item.shape) == 3:
            return ax.imshow(_draw_empty(f'allowed color-type in GRAY, RGB, RGBA, got {item.shape}'))
        return ax.imshow(Image.fromarray(item))

    if not bool(item):
        return ax.imshow(_draw_empty(f'item:{item}'))

    if isinstance(item, str):
        return ax.imshow(Image.open(item))

    if isinstance(item, Image.Image):
        return ax.imshow(item)

    if isinstance(item, List):
        # 点列类
        if len(item) > 2 and len(item[0]) == 2:
            # [(x, y)]
            xs, ys = zip(*item)
            return ax.plot(xs, ys)
        if len(item) == 2 and len(item[0]) > 2:
            # [xs, ys]
            xs, ys = item
            return ax.plot(xs, ys)
        shape = []
        while item and isinstance(item, (List, Tuple)) and len(item)>0:
            shape.append(len(item))
            item = item[0]
        return ax.imshow(_draw_empty(f'Only support formats like [(x, y)] or [xs, ys], must num > 2, got shape=={shape}'))

    if hasattr(item, 'geo'):
        item = item.geo

    if isinstance(item, BaseGeometry):
        l, u, r, d = list(map(float, item.bounds))
        ax.set_xticks([l, r])
        ax.set_yticks([u, d])
        m = max(l, u, r, d) * 0.05
        ax.set_xlim(l - m, r + m)
        ax.set_ylim(u - m, d + m)
        # ax.get_xaxis().set_visible(False)
        # ax.get_yaxis().set_visible(False)
        # 变换矩阵: matrix = [xAx, xAy, yAx, yAy, xb, yb]
        # img = affine_transform(img, [1, 0, 0, -1, 0, d])

        # 适配多图形和单图形
        geos = item.geoms if hasattr(item, 'geoms') else [item]
        for geo in geos:
            # 这是 shapely 的全部类型
            # [
            #     "Point",
            #     "LineString",
            #     "Polygon",
            #     "MultiPoint",
            #     "MultiLineString",
            #     "MultiPolygon",
            #     "GeometryCollection",
            #     "LinearRing",
            # ]
            if isinstance(geo, shapely.Point):
                ax.scatter(geo.x, geo.y)
                continue
            if isinstance(geo, (shapely.LineString, shapely.LinearRing)):
                xs, ys = geo.xy
                ax.plot(xs, ys)
                continue
            if isinstance(geo, shapely.Polygon):
                xs, ys = geo.exterior.xy
                ax.fill(xs, ys, color='blue', alpha=0.5)
                for interior in geo.interiors:
                    xs, ys = interior.xy
                    ax.fill(xs, ys, color='white', alpha=1)
                ax.set_aspect('equal')
                # ax.set_title('Polygon with Hole')
                # ax.legend()
                continue
            # 存在不兼容类型，直接清空之前画的东西
            ax.clear()
            return ax.imshow(_draw_empty(f'Support shapely type Point、LineString、LineRing、Polygon. Unknown with in-type: {type(geo)} -- {geo}'))
        return  # 显示 shapely
    return ax.imshow(_draw_empty(f'Unknown type: {type(item)} -- {item}'))


def _draw_empty(message: str) -> Image.Image:
    # 最大保留两百个字符，再多没意义
    message = message[:200]
    lines = [message[p * 30: (p + 1) * 30] for p in range(math.ceil(len(message) / 30))]
    temp = np.zeros((20 + 30 * len(lines), 20 + 19 * min(len(message), 30), 3), dtype=np.uint8)
    for p, txt in enumerate(lines):
        cv2.putText(temp, txt, (10, 10 + 30 * (p + 1)), cv2.FONT_HERSHEY_SIMPLEX, 1, [180, 255, 60])
    return Image.fromarray(temp)



from typing import List, Tuple
import random
import cv2
import numpy as np


def random_colors(
        n: int,
        start: Tuple[int, int, int] = (47, 81, 177),
        step: Tuple[int, int, int] = (73, 119, 31),
        rand: Tuple[int, int, int] = (10, 10, 10),
) -> List[Tuple[int, int, int]]:
    """
    生成一组随机色值，在一定程度上保证相邻色值间存在显著的用色差异，不关心色值是否存在复用现象
    :param n:       生成颜色数量
    :param start:   第一个颜色
    :param step:    颜色渐变公式（线性）
    :param rand:    随机因子
    :return:
    """
    r, g, b = start
    result = [(r, g, b)]
    while len(result) < n:
        r, g, b = [
            (
                c + step[i] + random.randint(-rand[i], +rand[i])
            ) % 255 for i, c in enumerate([r, g, b])
        ]
        result.append((r, g, b))
    return result


def random_rainbow_curves(
    shape: Tuple[int, int, int],
    s: int = 117,
    k: int = 7,
    c: int = 50,
):
    """
    生成一组随机色斑
    :param shape:   所生成的图像尺寸 (h, w, c)，按需取值
    :param s:       取值越大，色斑的尺寸越大，（影响速度），取值必须是奇数，推荐取值为图像尺寸的十分之一
    :param k:       迭代次数，取值越大，图案越平滑，（影响速度），推荐取值 0 - 20
    :param c:       对比度系数，取值越大，图案对比越强烈，推荐取值 0 - 100
    :return:
    """
    # 先生成噪声图
    h, w, nc = shape
    bg = np.random.random(w * h * nc).reshape((h, w, nc))
    bg = (np.clip(bg, 0, 1) * 256).astype(np.uint8)
    # 再将噪声图处理成背景干扰图像
    for _ in range(k):
        # 颜色空间拉伸，显眼化
        bg = bg.clip(c, 255 - c) - c
        bg = (bg / (255 - 2 * c) * 255).astype(np.uint8)
        # 用中值滤波对均值滤波做差，实现色斑合并
        m1 = cv2.blur(bg, (s, s))
        m2 = cv2.medianBlur(bg, s)
        bg = np.clip(2 * m2.astype(int) - m1.astype(int), 0, 255).astype(np.uint8)
    return bg
